#include "VT.h"

bool VT::asymptotic = false;
bool VT::permuteBool = false;
bool VT::vcfAnnotated = true;
bool VT::preProcess = false;
String VT::chopTo = "chop_to";
int VT::chop = 20;
int VT::permuteMin = 1000;
int VT::permuteMax = 3000000;
bool VT::VTwtMB = false;
bool VT::VTwtCMC_binary = false;
bool VT::VTwtCMC_counts = false;
bool VT::noStep = false;


VT::VT(){}

VT::~VT()
{}


bool VT::GetTmpGenoMatrixDosage(FastTransform & trans, Pedigree & ped,GroupFromAnnotation & group,int groupNum)
{
   double mac =0.0;
   bool status = false;
   tmpGenoMatrix.Dimension(0,0);
   mafIncluded.Dimension(0);

   int numSamples = header.getNumSamples();

   reader.set1BasedReadSection(group.chrList[groupNum][0].c_str(),group.posList[groupNum][0],group.posList[groupNum][group.posList[groupNum].Length()-1]+1);
   int m=0;
   while(reader.readRecord(record))
   {
      int position = record.get1BasedPosition();
      String altAllele = record.getAltStr();
      String refAllele = record.getRefStr();
      if(group.posList[groupNum][m]!=position)
      {
         continue;
      }
      else
      {
         String allele = refAllele + "/" + altAllele;
         if(allele != group.alleleList[groupNum][m])
         {
            String flipAllele = altAllele + "/" + refAllele;
            if(flipAllele != group.alleleList[groupNum][m])
               continue;
            else
            {
               //TODO:ref and alt are flipped. need to take that into account when calculating genotypes.
               //TODO: maybe good for t2dgenes, but not good for others
               printf("ref and alt alleles are flipped!");
            }
         }
      }
      //If the variant match the record in groupfile then add up m to next variant. 
      //This requires varaints in group file has to be sorted.
      m++;
      if(m==group.chrList[groupNum].Length())
         break;

      int n=0,nf=0;
      double fmaf = 0.0, maf=0.0;
      VcfRecordGenotype & genoInfo = record.getGenotypeInfo();

      //Loop througth all samples to calculate maf
      for(int s=0;s<numSamples;s++)
      {
         const std::string * geno = genoInfo.getString("DS",s);
         const char * sample = header.getSampleName(s);

         if(geno==NULL)
         {
            error("Error: Can not find DS field for sample %d ...",s);
         }
         if(*geno == ".")
         {
            continue;
         }
         float dose = atof((*geno).c_str());
         //judge if this is a founder
         int founder = trans.foundersHash.Integer(sample);
         if(founder != -1)
         {
            fmaf += dose;
            nf++;
         }
         maf += dose;
         n++;
      }
      if(n!=0)
         maf /= 2.0*n;
      else
         continue;

      if(nf!=0)
         fmaf /= 2.0*nf;
      else
         fmaf=maf;
      if(maf==0.0)
      {
         fmaf = 0.0;
      }
      if(fmaf==0.0 && maf>0)
         fmaf = maf;

      if(maf==0.0)
         continue;

      if(fmaf>0.5) fmaf = 1.0-fmaf;
      if(maf>0.5) maf = 1.0-maf;

      //check if maf falls into the range
      if(SVT::FounderFreq)
      {
         if(fmaf>mafMax || fmaf<mafMin)
            continue;
      }
      else
      {
         if(maf>mafMax || maf<mafMin)
            continue;
      }
      numVarPerGene++;
      if(fullResult)
      {
         String tmp = "";
         tmp += group.chrList[groupNum][m];
         tmp+=":";
         tmp+=group.posList[groupNum][m];
         tmp+= "_";
         tmp+=refAllele;
         tmp+= "/";
         tmp+=  altAllele;
         tmp+=  ";";
         varsIncluded += tmp;
      }

      if(SVT::FounderFreq)
      {
         mafIncluded.Push(fmaf);
      }
      else
      {
         mafIncluded.Push(maf);
      }

      tmpGenoMatrix.GrowTo(trans.persons,tmpGenoMatrix.cols+1,0.0);

      //update the genotype matrix
      int idx=0;
      for (int f = 0; f < ped.familyCount; f++)
      {
         for(int j=0;j<trans.pPheno[f].Length();j++)
         {
            String sample = ped[trans.pPheno[f][j]].pid.c_str();

            //get the sample number by sample ID from hash
            int s = trans.sampleVCFIDHash.Integer(sample);
            if(s!=-1)
            {
               const std::string * geno = genoInfo.getString("DS",s);
               if(strcmp((*geno).c_str(),".")>0)
               {
                  float dose = atof((*geno).c_str());
                  if(dose>0)
                  {
                     status=true;
                     if(CMC_binary)
                     {
                        tmpGenoMatrix[idx][numVarPerGene-1] = 1;
                     }
                     if(CMC_counts)
                     {
                        tmpGenoMatrix[idx][numVarPerGene-1] +=dose;
                     }
                     if(MB)
                     {
                        tmpGenoMatrix[idx][numVarPerGene-1] += dose/sqrt(mafIncluded[numVarPerGene-1]*(1.0-mafIncluded[numVarPerGene-1]));
                     }
                     mac += dose;
                  }
               }
            }
            idx++;
         }
      }
   }
   if(numVarPerGene==0)
   {
      status =false;
      return status;
   }
   if(mac<MAC)
{
      status = false;
return status;
 }
   //work on mafIncluded to get unique cutoffs
   Vector cp_mafIncluded;
   cp_mafIncluded.Copy(mafIncluded);
   cp_mafIncluded.Sort();
   unique_maf.Clear();
   unique_maf.Push(cp_mafIncluded[0]);
   for(int i=1;i<cp_mafIncluded.Length();i++)
   {
      if(cp_mafIncluded[i]>unique_maf[unique_maf.Length()-1])
         unique_maf.Push(cp_mafIncluded[i]);
   } //now unique maf cutoffs are saved in unique_maf.
  return status;
}
bool VT::GetTmpGenoMatrixVCF(FastTransform & trans, Pedigree & ped,GroupFromAnnotation & group,int groupNum)
{}

bool VT::GetTmpGenoMatrixPED(FastTransform & trans, Pedigree & ped,GroupFromAnnotation & group,int groupNum)
{}
bool VT::GetGenotypeMatrix(FastTransform & trans, Pedigree & ped,GroupFromAnnotation & group,int groupNum)
{
   bool status = false;

   if(SVT::genoFromPed)
      status = GetTmpGenoMatrixPED(trans,ped,group,groupNum);
   if(SVT::genoFromVCF)
      status = GetTmpGenoMatrixVCF(trans,ped,group,groupNum);
   if(SVT::dosage)
      status = GetTmpGenoMatrixDosage(trans,ped,group,groupNum);
   if(!status)
      return status;
}

bool VT::FinalizeGenoMatrix(FastTransform & trans,bool reduceDimension)
{
bool status = true;
   if(reduceDimension)
   {
      while(unique_maf.Length()>20)
      {
         unique_maf.Delete[0];
      }
   }

   genotypeMatrix.Dimension(trans.persons,unique_maf.Length());

   for(int j=0;j<unique_maf.Length();j++)
   {
      for(int i=0;i<mafIncluded.Length();i++)
      {
         if(mafIncluded[i]<unique_maf[j])
         {
            for(int p=0;p<trans.persons;p++)
            {
               genotypeMatrix[p][j] += tmpGenoMatrix[p][i];
            }
         }
      }

      if(VTwtCMC_binary)
      {
         for(int p=0;p<trans.persons;p++)
         {
            if(genotypeMatrix[p][j]>0.0)
               genotypeMatrix[p][j] = 1.0;
         }
      }
   }

 //adjusting mean
   for(int c=0;c<genotypeMatrix.cols;c++)
   {
      double mean =0.0;
      for(int r=0;r<trans.persons;r++)
      {
         mean += genotypeMatrix[r][c];
      }
      if(mean ==0.0)
         del.Push(c);
      else
      {
         mean /= trans.persons;
         for(int r=0;r<trans.persons;r++)
         {
            genotypeMatrix[r][c] -= mean;
         }
      }
   }
   if(del.Length()>0)
      for(int i=0;i<del.Length();i++)
      {
         genotypeMatrix.DeleteColumn(del[i]-i);
      }

if(genotypeMatrix.cols==0 || genotypeMatrix.rows==0.0)
status=false;
returne status;
}

void VT::CalculatePermute(Pedigree & ped, FastTransform & trans,FastFit & engine,String & traitName)
{
   FILE * VTout;

   String VTfilename;
   VTfilename = Burden::outputFile + traitName + ".VT.Permutation.Association.txt";
   VTout = fopen(VTfilename,"wt");

   int bad =0;
   if(Burden::genename == "" && Burden::genename_f == "")
   {
      bad = VTPermuteAll(ped,trans,engine,VTout,traitName);
   }
   else 
   {
      StringArray genes;
      //parse genename into string array
      if(Burden::genename_f!="")
      {
	 FILE * geneFile = fopen(Burden::genename_f,"r");
	 String buffer;
	 while (!feof(geneFile))
	 {
	    buffer.ReadLine(geneFile);
	    genes.AddTokens(buffer, SEPARATORS);
	 }
	 fclose(geneFile);
      } else if(Burden::genename != "") {
	 genes.AddTokens(Burden::genename,"/");
      }
      geneCount = genes.Length();

      bad = VTPermuteFile(ped,trans,engine,VTout,genes,traitName);
   }

   fprintf(VTout,"===================================================================\n");
   if(Burden::genename == "" && Burden::genename_f == "")
   {
      fprintf(VTout,"\n%d %s tested.\n",geneCount,geneCount>1?"genes were":"gene was");
      fprintf(VTout,"There %s %d %s either had no rare allele in the qualified sample or did not pass the collapsing criterion.\n",bad>1?"are":"is", bad,bad>1?"genes":"gene");
   } 
   else 
   {
      fprintf(VTout,"\n%d %s entered to be tested.\n",geneCount,geneCount>1?"genes were":"gene was");
      fprintf(VTout,"There %s %d %s either had no rare allele in the qualified sample or did not pass the collapsing criterion, or was not found in the annotated .vcf file.\n",bad>1?"are":"is", bad,bad>1?"genes":"gene");
   }
   fclose(VTout);
}

int VT::VTPermuteAll(Pedigree & ped, FastTransform & trans, FastFit & engine,FILE * VTout,String & traitName)
{
   time_t ti;
   Random *rand = new Random((long)time(&ti));
   double sigma_g2 = engine.sigma_g2Hat;
   //double sigma_e2 = engine.sigma_e2Hat;
   double h2 = sigma_g2/trans.traitVar;

   if(VTwtMB + VTwtCMC_binary + VTwtCMC_counts ==1)
   {
      if(VTwtMB)
      {
	 fprintf(VTout,"\nWeight used is Madsen-Browning.\n");
      } 
      else if(VTwtCMC_binary)
      {
	 fprintf(VTout,"\nWeight used is CMC_binary.\n");
      } 
      else
      {
	 fprintf(VTout,"\nWeight used is CMC_counts.\n");
      }
   }
   else 
   {
      fprintf(VTout,"\nWeights used are: %s%s%s.\n",VTwtMB?"Madsen-Browning ":"",VTwtCMC_binary?"CMC_binary ":"",VTwtCMC_counts?"CMC_counts":"");
   }

   fprintf(VTout,"\n%s mean=%6.2f, variance=%6.2f, heritability=%6.2f \n",traitName.c_str(),trans.traitMean,trans.traitVar,h2*100.0);
   fprintf(VTout,"===================================================================\n");
   fprintf(VTout,"gene\tpermutation_pvalue\n");

   int bad =0;
   int gene_counter =0;
   for(int gene=0;gene<geneHash.Capacity();gene++)
   {
      if(!geneHash.SlotInUse(gene)) 
      {
	 continue;
      }

      String geneName = geneHash[gene];
      int value = geneHash.Integer(geneName);

      if(geneName.IsEmpty())
      {
	 continue;
      }
      gene_counter++;
      printf("Testing gene %s.\n",geneName.c_str());

      if(GetGenotypeMatrix(trans,ped,geneName,geneChrom[value],start[value],end[value]+1,true))
      {
	 Matrix transGeno;
	 transGeno.Product(trans.UD,genotypeMatrix);

	 Vector score,permutedResidual,Y_star,permutedBeta,permutedFixedEff;
	 score.Dimension(transGeno.cols);
	 permutedResidual.Copy(residuals_original);
	 permutedBeta.Dimension(trans.fixedEffects);
	 permutedFixedEff.Dimension(trans.persons);
	 double permutedSigma2,numerator,score_original;

	 Vector denominator_2,denominator;
	 denominator_2.Dimension(trans.fixedEffects);
	 denominator.Dimension(transGeno.cols);

	 //Get the constants across all permutations
	 for(int cutoff=0;cutoff<transGeno.cols;cutoff++)
	 {
	    denominator[cutoff] = 1.0e-10;
	    denominator_2.Zero();
	    for(int i=0;i<trans.persons;i++)
	    {
	       denominator += transGeno[i][cutoff]*transGeno[i][cutoff];
	       for(int j=0;j<trans.fixedEffects;j++)
	       {
		  denominator_2[j] += transGeno[i][cutoff]*trans.UDX[i][j];
	       }
	    }
	    Vector tmp; //tmp is T(sum_SiZi)*inv
	    tmp.Dimension(trans.fixedEffects);
	    for(int i=0;i<trans.fixedEffects;i++)
	    {
	       tmp[i] = trans.inv[i].InnerProduct(denominator_2);
	    }
	    denominator[cutoff] -= tmp.InnerProduct(denominator_2);
	 }

	 //Calculate original T_score
	 for(int cutoff=0;cutoff<transGeno.cols;cutoff++)
	 {
	    numerator = 0.0;
	    for(int i=0;i<trans.persons;i++)
	    {
	       numerator += residuals_original[i]*transGeno[i][cutoff];
	    }

	    if(denominator==0) score[cutoff]=0.0;
	    else score[cutoff] = fabs(numerator/sqrt(denominator[cutoff]));
	    // effSize = -numerator/denominator;
	 }
	 score_original = score.Max()/sqrt(sigma2_hat);

	 //start permutations
	 double pvalue;
	 int count_ =0;
	 int permute;
	 Vector permutedResidual_;
	 permutedResidual_.Dimension(trans.persons);

	 for(int p=0;p<permuteMin;p++)
	 {
	    //permute the trait
	    for(int r=0;r<trans.persons;r++)
	    {
	       int randNum = (int)(rand->Uniform(0.0,trans.persons));
	       while(randNum == r)
	       {
		  randNum = (int)(rand->Uniform(0.0,trans.persons));
	       }
	       permutedResidual.Swap(r,randNum);
	    }
	    //Reconstruct trait values
	    Y_star.Copy(permutedResidual);
	    Y_star.Add(fixedEff);
	    engine.EstimateAnalyticalBeta(Y_star,trans,permutedBeta);
	    for(int i=0;i<trans.persons;i++)
	    {
	       permutedFixedEff[i] = trans.UDX[i].InnerProduct(permutedBeta);
	    }
	    //Calculate new residuals
	    for(int i=0;i<trans.persons;i++)
	       permutedResidual_[i] = trans.UDY[i] - permutedFixedEff[i];
	    permutedSigma2 = permutedResidual_.SumSquares()/trans.persons;

	    //Calculate the statistics
	    for(int cutoff=0;cutoff<transGeno.cols;cutoff++)
	    {
	       numerator = 0.0;
	       for(int i=0;i<trans.persons;i++)
	       {
		  numerator += permutedResidual_[i]*transGeno[i][cutoff];
	       }
	       if(denominator[cutoff]==0) score[cutoff]=0.0;
	       else score[cutoff] = fabs(numerator/sqrt(permutedSigma2*denominator[cutoff]));
	    }
	    if(score_original < score.Max())
	    {
	       count_++;
	    }
	 }
	 for(int p=permuteMin;p<permuteMax;p++)
	 {
	    //permute the trait
	    for(int r=0;r<trans.persons;r++)
	    {
	       int randNum = (int)(rand->Uniform(0.0,trans.persons));
	       while(randNum == r)
	       {
		  randNum = (int)(rand->Uniform(0.0,trans.persons));
	       }
	       permutedResidual.Swap(r,randNum);
	    }
	    //Reconstruct trait values
	    Y_star.Copy(permutedResidual);
	    Y_star.Add(fixedEff);
	    engine.EstimateAnalyticalBeta(Y_star,trans,permutedBeta);
	    for(int i=0;i<trans.persons;i++)
	    {
	       permutedFixedEff[i] = trans.UDX[i].InnerProduct(permutedBeta);
	    }
	    //Calculate new residuals
	    for(int i=0;i<trans.persons;i++)
	       permutedResidual_[i] = trans.UDY[i] - permutedFixedEff[i];
	    permutedSigma2 = permutedResidual_.SumSquares()/trans.persons;

	    //Calculate the statistics
	    for(int cutoff=0;cutoff<transGeno.cols;cutoff++)
	    {
	       numerator = 0.0;
	       for(int i=0;i<trans.persons;i++)
	       {
		  numerator += permutedResidual_[i]*transGeno[i][cutoff];
	       }
	       if(denominator[cutoff]==0) score[cutoff]=0.0;
	       else score[cutoff] = fabs(numerator/sqrt(permutedSigma2*denominator[cutoff]));
	    }
	    if(score_original < score.Max())
	    {
	       count_++;
	    }
	    permute = p+1;
	    if(noStep)
	    {
	       continue;
	    }
	    if(count_ >= 50)
	    {
	       break;
	    }
	 }
	 if(count_==0)
	    fprintf(VTout,"%s\t<%g\n",geneName.c_str(),1.0/permute);
	 else
	 {
	    pvalue  = (double)count_/permute;
	    fprintf(VTout,"%s\t%g\n",geneName.c_str(),pvalue);
	 } 
      } 
      else 
      {
	 bad++;
	 fprintf(VTout,"%s\t-\n",geneName.c_str());
      }
   }
   if(rand) delete rand;
   return bad;
}

int VT::VTPermuteFile(Pedigree & ped, FastTransform & trans, FastFit & engine,FILE * VTout,StringArray & genes,String & traitName)
{
   time_t ti;
   Random *rand = new Random((long)time(&ti));

   double sigma_g2 = engine.sigma_g2Hat;
   double sigma_e2 = engine.sigma_e2Hat;
   double h2 = sigma_g2/(sigma_g2+sigma_e2);

   if(VTwtMB + VTwtCMC_binary + VTwtCMC_counts ==1)
   {
      if(VTwtMB)
	 fprintf(VTout,"\nWeight used is Madsen-Browning.\n");
      else if(VTwtCMC_binary)
	 fprintf(VTout,"\nWeight used is CMC_binary.\n");
      else
	 fprintf(VTout,"\nWeight used is CMC_binary.\n");
   }
   else
   {
      fprintf(VTout,"\nWeights used are: %s%s%s.\n",VTwtMB?"Madsen-Browning ":"",VTwtCMC_binary?"CMC_binary ":"",VTwtCMC_counts?"CMC_counts":"");
   }

   fprintf(VTout,"\n%s mean=%6.2f, variance=%6.2f, heritability=%6.2f \n",traitName.c_str(),trans.traitMean,trans.traitVar,h2*100.0);
   fprintf(VTout,"===================================================================\n");
   fprintf(VTout,"gene\tpermutation_pvalue\n");

   int bad =0;
   int gene_counter=0;
   for(int g=0;g<genes.Length();g++)
   {
      int value = geneHash.Integer(genes[g]);

      if(value ==-1)
      {
	 fprintf(VTout,"gene %s is not found in the annotated file.\n",genes[g].c_str());
	 bad++;
	 continue;
      }
      gene_counter++;
      printf("Testing gene %s.\n",genes[g].c_str());


      if(GetGenotypeMatrix(trans,ped,genes[g],geneChrom[value],start[value],end[value]+1,true))
      {
	 Matrix transGeno;
	 transGeno.Product(trans.UD,genotypeMatrix);

	 Vector score,permutedResidual,Y_star,permutedBeta,permutedFixedEff;
	 score.Dimension(transGeno.cols);
	 permutedResidual.Copy(residuals_original);
	 permutedBeta.Dimension(trans.fixedEffects);
	 permutedFixedEff.Dimension(trans.persons);
	 double permutedSigma2,numerator,score_original;

	 Vector denominator_2,denominator;
	 denominator_2.Dimension(trans.fixedEffects);
	 denominator.Dimension(transGeno.cols);

	 //Get the constants across all permutations
	 for(int cutoff=0;cutoff<transGeno.cols;cutoff++)
	 {
	    denominator[cutoff] = 1.0e-10;
	    denominator_2.Zero();
	    for(int i=0;i<trans.persons;i++)
	    {
	       denominator += transGeno[i][cutoff]*transGeno[i][cutoff];
	       for(int j=0;j<trans.fixedEffects;j++)
	       {
		  denominator_2[j] += transGeno[i][cutoff]*trans.UDX[i][j];
	       }
	    }
	    Vector tmp; //tmp is T(sum_SiZi)*inv
	    tmp.Dimension(trans.fixedEffects);
	    for(int i=0;i<trans.fixedEffects;i++)
	    {
	       tmp[i] = trans.inv[i].InnerProduct(denominator_2);
	    }
	    denominator[cutoff] -= tmp.InnerProduct(denominator_2);
	 }

	 //Calculate original T_score
	 for(int cutoff=0;cutoff<transGeno.cols;cutoff++)
	 {
	    numerator = 0.0;
	    for(int i=0;i<trans.persons;i++)
	    {
	       numerator += residuals_original[i]*transGeno[i][cutoff];
	    }

	    if(denominator==0) score[cutoff]=0.0;
	    else score[cutoff] = fabs(numerator/sqrt(denominator[cutoff]));
	    // effSize = -numerator/denominator;
	 }
	 score_original = score.Max()/sqrt(sigma2_hat);

	 //start permutations
	 int count_ =0;
	 double pvalue;
	 int permute;
	 Vector permutedResidual_;
	 permutedResidual_.Dimension(trans.persons);

	 for(int p=0;p<permuteMin;p++)
	 {
	    //permute the trait
	    for(int r=0;r<trans.persons;r++)
	    {
	       int randNum = (int)(rand->Uniform(0.0,trans.persons));
	       while(randNum == r)
	       {
		  randNum = (int)(rand->Uniform(0.0,trans.persons));
	       }
	       permutedResidual.Swap(r,randNum);
	    }
	    //Reconstruct trait values
	    Y_star.Copy(permutedResidual);
	    Y_star.Add(fixedEff);
	    engine.EstimateAnalyticalBeta(Y_star,trans,permutedBeta);
	    for(int i=0;i<trans.persons;i++)
	    {
	       permutedFixedEff[i] = trans.UDX[i].InnerProduct(permutedBeta);
	    }
	    for(int i=0;i<trans.persons;i++)
	       permutedResidual_[i] = trans.UDY[i] - permutedFixedEff[i];
	    permutedSigma2 = permutedResidual_.SumSquares()/trans.persons;

	    //Calculate the statistics
	    for(int cutoff=0;cutoff<transGeno.cols;cutoff++)
	    {
	       numerator = 0.0;
	       for(int i=0;i<trans.persons;i++)
	       {
		  numerator += permutedResidual_[i]*transGeno[i][cutoff];
	       }
	       if(denominator[cutoff]==0) score[cutoff]=0.0;
	       else score[cutoff] = fabs(numerator/sqrt(permutedSigma2*denominator[cutoff]));
	    }
	    if(score_original < score.Max())
	    {
	       count_++;
	    }
	 }
	 for(int p=permuteMin;p<permuteMax;p++)
	 {
	    //permute the trait
	    for(int r=0;r<trans.persons;r++)
	    {
	       int randNum = (int)(rand->Uniform(0.0,trans.persons));
	       while(randNum == r)
	       {
		  randNum = (int)(rand->Uniform(0.0,trans.persons));
	       }
	       permutedResidual.Swap(r,randNum);
	    }
	    //Reconstruct trait values
	    Y_star.Copy(permutedResidual);
	    Y_star.Add(fixedEff);
	    engine.EstimateAnalyticalBeta(Y_star,trans,permutedBeta);
	    for(int i=0;i<trans.persons;i++)
	    {
	       permutedFixedEff[i] = trans.UDX[i].InnerProduct(permutedBeta);
	    }
	    for(int i=0;i<trans.persons;i++)
	       permutedResidual_[i] = trans.UDY[i] - permutedFixedEff[i];
	    permutedSigma2 = permutedResidual_.SumSquares()/trans.persons;

	    //Calculate the statistics
	    for(int cutoff=0;cutoff<transGeno.cols;cutoff++)
	    {
	       numerator = 0.0;
	       for(int i=0;i<trans.persons;i++)
	       {
		  numerator += permutedResidual_[i]*transGeno[i][cutoff];
	       }
	       if(denominator[cutoff]==0) score[cutoff]=0.0;
	       else score[cutoff] = fabs(numerator/sqrt(permutedSigma2*denominator[cutoff]));
	    }
	    if(score_original < score.Max())
	    {
	       count_++;
	    }
	    permute = p +1;
	    if(noStep)
	       continue;
	    if(count_ >= 50)
	    {
	       break;
	    }
	 }
	 if(count_==0)
	    fprintf(VTout,"%s\t<%g\n",genes[g].c_str(),1.0/permute);
	 else
	 {
	    pvalue  = (double)count_/permute;
	    fprintf(VTout,"%s\t%g\n",genes[g].c_str(),pvalue);
	 } 
      }
      else
      {  
	 bad++;
	 fprintf(VTout,"%s\t-\n",genes[g].c_str());
      }
   }
   if(rand) delete rand;
   return bad;
}

void VT::ChopAndWrite(Pedigree & ped)
{
   GetAllGenenamesVCF(Burden::chromosome);
   reader.open(GroupFromAnnotation::vcfInput,header);
   reader.readVcfIndex();
   StringArray SNPNameList,outputVCF;
   IntArray posList,chromList;

   //Make directory to save the chopped files
   String command;
   command = "rm -r " + chopTo;
   system(command.c_str());
   command = "mkdir " + chopTo;
   system(command.c_str());

   int fileSize;
   fileSize = (int) geneCount/chop + 1;
   printf("Chopping input files into %d %s and saving them to /%s ...\n",fileSize,fileSize>1?"files":"file",chopTo.c_str());

   String vcfName,pedName,datName;

   FILE * vcfOut;
   FILE * pedOut;
   FILE * datOut;

   int gene_counter =0;
   int idx = 0;
   if(Burden::genename == "" && Burden::genename_f == "")
   {
      for(int i=0;i<geneHash.Capacity();i++)
      {
	 if(!geneHash.SlotInUse(i))
	 {
	    continue;
	 }
	 String geneName = geneHash[i];
	 int value = geneHash.Integer(geneName);

	 if(geneName.IsEmpty())
	 {
	    continue;
	 }

	 if(gene_counter % chop ==0)
	 {
	    outputVCF.Clear();
	    chromList.Clear();
	    posList.Clear();
	    SNPNameList.Clear();
	    idx++;

	    if(chopTo[chopTo.Length()-1]=='/')
	    {
	       vcfName = chopTo + idx + ".vcf";
	       pedName = chopTo + idx + ".ped";
	       datName = chopTo + idx + ".dat";
	    }
	    else
	    {
	       vcfName = chopTo + "/"+ idx + ".vcf";
	       pedName = chopTo + "/"  + idx + ".ped";
	       datName = chopTo + "/" + idx + ".dat";
	    }

	    vcfOut = fopen(vcfName,"wt");
	    fprintf(vcfOut,"#CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\n");
	    pedOut = fopen(pedName,"wt");
	    datOut = fopen(datName,"wt");
	 }

	 //Read vcf file to find the gene and then output
	 const std::string* txGNstr;
	 const std::string* FUNCstr;
	 reader.set1BasedReadSection(geneChrom[value].c_str(),start[value],end[value]+1);

	 bool found_gene=false;

	 while(reader.readRecord(record))
	 {
	    VcfRecordInfo & info = record.getInfo();
	    txGNstr =  info.getString("txGN");
	    FUNCstr =  info.getString("FUNC");
	    if(txGNstr == NULL)
	    {
	       continue;
	    }

	    size_t found = txGNstr->find(geneName.c_str());
	    if(found == std::string::npos)
	    {
	       continue;
	    }
	    /* bug here
	       if(txGNstr->length()>geneName.Length())
	       {
	       int newpos = found+geneName.Length();
	       if((*txGNstr)[newpos]!='/'  && (*txGNstr)[newpos] != '\r' && (*txGNstr)[newpos] !='\n' && (*txGNstr)[newpos] !='\0')
	       {
	    //geneName is only a substring of another gene's name
	    continue;
	    }
	    }
	     */
	    if(txGNstr->length()==geneName.Length())
	    {
	       found_gene = true;
	    } 
	    else if(txGNstr->length()>geneName.Length())
	    {  
	       StringArray tmp;
	       tmp.AddTokens((*txGNstr).c_str(),SEPARATORS);
	       for(int i=0;i<tmp.Length();i++)
	       {
		  if(tmp[i].Length()!=geneName.Length())
		     continue;
		  else 
		  {
		     if(tmp[i].FastCompare(geneName)==0)	
		     {
			found_gene=true;
			break;
		     }
		     else
		     {
			continue;
		     }
		  }
	       }
	       if(!found_gene) continue;
	    }

	    //output the record to vcf file

	    const char* tmp_chrom = record.getChromStr();
	    int pos = record.get1BasedPosition();

	    String tmp_vcf;
	    tmp_vcf.Copy(tmp_chrom);
	    tmp_vcf += '\t';
	    tmp_vcf += record.get1BasedPosition();
	    tmp_vcf += '\t'; 
	    tmp_vcf += record.getIDStr();
	    tmp_vcf +=  '\t';
	    tmp_vcf += record.getRefStr();
	    tmp_vcf += '\t';
	    tmp_vcf += record.getAltStr();
	    tmp_vcf += '\t';
	    tmp_vcf += record.getQualStr();
	    tmp_vcf +=  '\t';
	    tmp_vcf +=  record.passedAllFilters()?"PASS":"FAIL";
	    tmp_vcf +=  "\ttxGN=";
	    tmp_vcf += geneName;
	    tmp_vcf +=  ";FUNC=";
	    tmp_vcf += (*FUNCstr).c_str();
	    //printf("tmp_vcf is:%s\n",tmp_vcf.c_str());

	    String SNP;
	    String tmpStr;
	    tmpStr.Copy(record.getChromStr());
	    int found_ = tmpStr.FastFindChar('r',0);
	    int insert_pos;
	    int chrom_;
	    if(found_ ==-1)
	    {
	       SNP = "chr";
	       SNP += tmpStr;
	       SNP += ":";
	       SNP+= pos;
	       //compare with the previous chromosome
	       chrom_ = atoi(tmp_chrom);
	    } 
	    else
	    {
	       SNP.Copy(tmpStr);
	       SNP += ":";
	       SNP+= pos;
	       chrom_ = atoi(tmpStr.SubStr(found+1));
	    }

	    if(gene_counter % chop ==0)
	    {
	       chromList.Push(chrom_);
	       posList.Push(pos);
	       //markerID.Push(ped.GetMarkerID(SNP));
	       SNPNameList.Push(SNP);
	       outputVCF.Push(tmp_vcf);
	    }
	    else 
	    {
	       //Decide where to insert the new record in Array
	       if(chrom_ >= chromList[chromList.Length()-1])
	       {
		  insert_pos = chromList.Length();
	       }
	       else
	       {
		  for(int i=0;i<chromList.Length();i++)
		  {
		     if(chrom_ >= chromList[i])
		     {
			continue;
		     }
		     insert_pos = i;
		     break;
		  }
	       }

	       for(int i=0;i<insert_pos;i++)
	       {
		  if(chrom_ > chromList[i])
		  {
		     continue;
		  }
		  if(chrom_== chromList[i])
		  {
		     if(pos > posList[i])
			continue;	
		     insert_pos = i;
		     break;
		  }
	       }

	       if(insert_pos == chromList.Length())
	       {
		  chromList.Push(chrom_);
		  posList.Push(pos);
		  SNPNameList.Push(SNP);
		  outputVCF.Push(tmp_vcf);
	       }
	       else
	       {
		  chromList.InsertAt(insert_pos,chrom_);
		  posList.InsertAt(insert_pos,pos);
		  SNPNameList.InsertAt(insert_pos,SNP);
		  outputVCF.InsertAt(insert_pos,tmp_vcf);
	       }
	    }
	 }
	 if(idx==20)
	    printf("gene %s is %s\n",geneName.c_str(),found_gene ? "found":"not found");

	 //Output ped file and dat file
	 if(gene_counter % chop == chop-1 || gene_counter == geneCount -1)
	 {
	    //  printf("gene_counter is: %d; chop is: %d.\n",gene_counter,chop);
	    //	    printf("SNPNameList length is:%d; outputVCF length: %d\n",SNPNameList.Length(),outputVCF.Length());

	    //output vcf file and  dat file
	    for(int i=0;i<SNPNameList.Length();i++)
	    {
	       fprintf(datOut,"M %s\n",SNPNameList[i].c_str());
	       fprintf(vcfOut,"%s\n",outputVCF[i].c_str());
	    }
	    for (int j = 0; j < ped.covariateCount; j++)
	    {
	       fprintf(datOut,"C %s\n",ped.covariateNames[j].c_str());
	    }
	    for(int j=0;j<ped.traitCount;j++)
	    {
	       fprintf(datOut,"T %s\n",ped.traitNames[j].c_str());
	    }

	    //output ped file. 
	    MarkerInfo * info;
	    int markerID;
	    for(int i=0;i<ped.count;i++)
	    {
	       fprintf(pedOut,"%s %s %s %s %d",ped.persons[i]->famid.c_str(),ped.persons[i]->pid.c_str(),ped.persons[i]->motid.c_str(),ped.persons[i]->fatid.c_str(),ped.persons[i]->sex);

	       for(int m=0;m<SNPNameList.Length();m++)
	       {
		  markerID = ped.GetMarkerID(SNPNameList[m]);
		  info = ped.GetMarkerInfo(markerID);
		  fprintf(pedOut," %s/%s",info->GetAlleleLabel(ped.persons[i]->markers[markerID][0]).c_str(),info->GetAlleleLabel(ped.persons[i]->markers[markerID][1]).c_str());
	       }

	       //print covariates and phenotypes
	       for(int j=0;j<ped.covariateCount;j++)
	       {
		  double tmp = ped.persons[i]->covariates[j];
		  if(tmp == _NAN_)
		  {
		     fprintf(pedOut," x");
		  }
		  else
		  {
		     fprintf(pedOut," %g",tmp);
		  }
	       }
	       for(int j=0;j<ped.traitCount;j++)
	       {
		  double tr = ped.persons[i]->traits[j];
		  if(tr == _NAN_)
		  {
		     fprintf(pedOut," x");
		  }
		  else
		  {
		     fprintf(pedOut," %g",tr);
		  }
	       }
	       fprintf(pedOut,"\n");
	    }

	    fclose(vcfOut);
	    fclose(pedOut);
	    fclose(datOut);

	    //Need to bgzip and tabix the vcf file here.
	    String command;
	    command = "bgzip " + vcfName;
	    system(command);
	    command = "tabix -p vcf " + vcfName + ".gz";
	    system(command);
	 }
	 gene_counter++;
      }
   }

   reader.close();
}

void VT::CalculateAsymptotic(Pedigree & ped, FastTransform & trans,FastFit & engine,String & traitName,GroupFromAnnotation & group)
{
   //double sigma_g2 = engine.sigma_g2Hat;
   //double sigma_e2 = engine.sigma_e2Hat;
   //double h2 = sigma_g2/(sigma_g2+sigma_e2);
  // printf("g,e,var:%g,%g,%g\n",sigma_g2,sigma_e2,trans.traitVar);
   double h2 = engine.sigma_g2Hat/trans.traitVar;

   FILE * VTout;
   FILE * VTlog;

   String VTfilename;
   VTfilename = Burden::outputFile + traitName + ".VT.Association.txt";
   VTout = fopen(VTfilename,"wt");
   VTfilename = Burden::outputFile + traitName + ".VT.log";
   VTlog = fopen(VTfilename,"wt");

   if(VTwtMB + VTwtCMC_binary + VTwtCMC_counts ==1)
   {
      if(VTwtMB)
      {
	 fprintf(VTout,"\nWeight used is Madsen-Browning.\n");
	 fprintf(VTlog,"\nWeight used is Madsen-Browning.\n");
      } 
      else if(VTwtCMC_binary)
      {
	 fprintf(VTout,"\nWeight used is CMC_binary.\n");
	 fprintf(VTlog,"\nWeight used is CMC_binary.\n");
      } 
      else
      {
	 fprintf(VTout,"\nWeight used is CMC_counts.\n");
	 fprintf(VTlog,"\nWeight used is CMC_counts.\n");
      } 
   }
   else
   {
      fprintf(VTout,"\nWeights used are: %s%s%s.\n",VTwtMB?"Madsen-Browning ":"",VTwtCMC_binary?"CMC_binary ":"",VTwtCMC_counts?"CMC_counts ":"");
      fprintf(VTlog,"\nWeights used are: %s%s%s.\n",VTwtMB?"Madsen-Browning ":"",VTwtCMC_binary?"CMC_binary ":"",VTwtCMC_counts?"CMC_counts ":"");
   }

   fprintf(VTout,"\n%s mean=%6.2f, variance=%6.2f, heritability=%6.2f \n",traitName.c_str(),trans.traitMean,trans.traitVar,h2*100.0);
   fprintf(VTout,"===================================================================\n");
   fprintf(VTlog,"===================================================================\n");
   fprintf(VTout,"gene\tasymptotic_pvalue\terror\n");
   fprintf(VTlog,"gene\tmethod\tcutoff_to_choose_from\tselected_cutoff\n");

   int bad =0;
   Matrix rh;
   Matrix UDX_t,tmp;
   UDX_t.Transpose(trans.UDX);
   tmp.Product(trans.inv,UDX_t);
   rh.Transpose(tmp);

   for(int i=0;i<group.annoGroups.Length();i++)
   {
      String geneName = group.annoGroups[i];
      printf("Testing gene %s.\n",geneName.c_str());

      if(GetGenotypeMatrix(trans,ped,group,i))
      {
	 if(!FinalizeGenoMatrix(trans,false))
	 {
	    //print output here
	    bad++;
	    fprintf(VTout,"%s\t-\t-\n",geneName.c_str());
	    fprintf(VTlog,"%s\t-\t-\t-\n",geneName.c_str());
	    continue;
	 }

      }
      else 
      {
	 //print output here
	 bad++;
	 fprintf(VTout,"%s\t-\t-\n",geneName.c_str());
	 fprintf(VTlog,"%s\t-\t-\t-\n",geneName.c_str());
	 continue;
      }
      Matrix transGeno;	
      transGeno.Product(trans.UD,genotypeMatrix);

      Vector score;
      score.Dimension(transGeno.cols);
      double pvalue,numerator,denominator;
      Vector denominator_2;
      denominator_2.Dimension(trans.fixedEffects);

      Matrix U;
      U.Dimension(transGeno.cols,trans.persons);

      for(int cutoff=0;cutoff<transGeno.cols;cutoff++)
      {
	 numerator = 0.0;
	 denominator = 1.0e-10;
	 denominator_2.Zero();
	 for(int i=0;i<trans.persons;i++)
	 {
	    numerator += residuals_original[i]*transGeno[i][cutoff];
	    denominator += transGeno[i][cutoff]*transGeno[i][cutoff];
	    for(int j=0;j<trans.fixedEffects;j++)
	    {
	       denominator_2[j] += transGeno[i][cutoff]*trans.UDX[i][j];
	    } 
	 }
	 Vector tmp; //tmp is T(sum_SiZi)*inv
	 tmp.Dimension(trans.fixedEffects);

	 for(int i=0;i<trans.fixedEffects;i++)
	 {
	    tmp[i] = trans.inv[i].InnerProduct(denominator_2);
	 }

	 denominator -= tmp.InnerProduct(denominator_2); 
	 //fill in U_ki for covariance matrix
	 for(int p=0;p<trans.persons;p++)
	 {
	    U[cutoff][p] = residuals_original[p] *(transGeno[p][cutoff] - tmp.InnerProduct(trans.UDX[p]));
	 }

	 if(denominator==0) score[cutoff]=0.0;
	 else score[cutoff] = fabs(numerator/sqrt(denominator))/sqrt(sigma2_hat);
	 // effSize = -numerator/denominator;
      }

      int cutoff = transGeno.cols;
      //get the chosen cutoff and output to .log file
      double t_max = score[0];
      int chosen_cutoff = 0;
      for(int i=1;i<cutoff;i++)
      {
	 if(t_max < score[i])
	 {
	    t_max = score[i];
	    chosen_cutoff = i;
	 }
      }
      //double t_max = score.Max()/sqrt(sigma2_hat);

      //output to .log file
      String method;
      if(VTwtMB + VTwtCMC_binary + VTwtCMC_counts ==1)
      {
	 if(VTwtMB) 
	    method = "MB";
	 if(VTwtCMC_binary)
	    method = "CMC_binary";
	 if(VTwtCMC_counts)
	    method = "CMC_counts";
      }
      else if(VTwtMB && VTwtCMC_binary && !VTwtCMC_counts)
      {
	 if(chosen_cutoff < length)
	    method = "MB";
	 else 
	    method = "CMC_binary";
      }
      else if(VTwtMB && VTwtCMC_counts && !VTwtCMC_binary)
      {
	 if(chosen_cutoff < length)
	    method = "MB";
	 else
	    method = "CMC_counts";
      }
      else if(VTwtCMC_counts && VTwtCMC_binary && !VTwtMB)
      {
	 if(chosen_cutoff < length)
	    method = "CMC_binary";
	 else
	    method = "CMC_counts";
      }
      else if(VTwtMB + VTwtCMC_binary + VTwtCMC_counts ==3)
      {
	 if(chosen_cutoff < length)
	    method = "MB";
	 else if(chosen_cutoff < 2*length)
	    method = "CMC_binary";
	 else 
	    method = "CMC_counts";
      }

      //	    printf("chosen_cutoff is:%d\n",chosen_cutoff);
      double selected_maf = unique_maf[chosen_cutoff];
      fprintf(VTlog,"%s\t%s\t%d\t%g\n",geneName.c_str(),method.c_str(),score.Length(),selected_maf);

      double * lower = new double [cutoff];
      double * upper = new double [cutoff];
      double * mean = new double [cutoff];

      for(int i=0;i<cutoff;i++)
      {
	 mean[i] = 0.0;
	 lower[i] = -t_max;
	 upper[i] = t_max;
      }

      Matrix cov_U; //this is the cov matrix for U
      cov_U.Dimension(cutoff,cutoff);
      for(int r=0;r<cutoff;r++)
      {
	 for(int c=r;c<cutoff;c++)
	 {
	    cov_U[r][c] = cov_U[c][r] = U[r].InnerProduct(U[c]);
	 }
      }

      Matrix cov; //this is the cov matrix for T
      cov.Dimension(cov_U.rows,cov_U.cols);
      cov2cor(cov_U,cov);

      //Use pmvnorm to calculate the asymptotic p-value
      Vector result;
      result.Copy(pmvnorm(lower,upper,mean,cov,false));
      //printf("pmvnorm results:%g,%g,%g\n",result[0],result[1],result[2]);
      if( 1.0 - result[0] ==0.0)
      {
	 printf("using Shuang's algorithm for mvnorm prob ...\n");
	 //printf("result0 is:%g\n",result[0]);
//TODO:reduce dimension here
/*
	 if(unique_maf.Length()>20)
	    FinalizeGenoMatrix(trans,true);
*/
	 pvalue = CalculatePvalue(score,cov,t_max);
	 fprintf(VTout,"%s\t%g\n",geneName.c_str(),pvalue);
      }
else if(result[0]=-1.0)
{
 fprintf(VTout,"%s\tERROR:CORR_MATRIX_NOT_POS_SEMI_DEF\n",geneName.c_str());
}
      else
      {
	 pvalue = 1.0 - result[0];
	 fprintf(VTout,"%s\t%g\t%g\n",geneName.c_str(),pvalue,result[1]);
      }

      if(lower) delete [] lower;
      if(upper) delete [] upper;
      if(mean) delete [] mean;
   }

   fprintf(VTout,"===================================================================\n");
   fprintf(VTout,"\n%d %s tested.\n",geneCount,geneCount>1?"genes were":"gene was");
   fprintf(VTout,"There %s %d %s either had no rare allele in the qualified sample or did not pass the collapsing criterion.\n",bad>1?"are":"is", bad,bad>1?"genes":"gene");

   fclose(VTout);
   fclose(VTlog);
}

double VT::CalculatePvalue(Vector & score,Matrix & cov,double t_max)
{
   double pvalue=0.0;
   int dim = score.Length();
   double chisq = t_max * t_max;

   StringDoubleHash jointProbHash;

   if(dim==1)
   {
      pvalue = chidist(chisq,1);
      //printf("pvalue is: %g\n",pvalue);
      return pvalue;
   }
   double uni = chidist(chisq,1);
   pvalue += dim*uni;
   //printf("pvalue so far is:%g\n",pvalue);

   //Initializations for combination
   vector<int> indx;
   vector<int> alpha (dim,0);
   for(int i=0;i<dim;i++)
      alpha[i] = i;
   int n=dim;

   for(int r=2;r<=dim;r++)
   {
      int j=r;
      int k=r;

      Vector par,comb;
      par.Dimension(2);

      for(int twk=j;twk<=k;twk++)
      {
	 int r=twk;
	 bool done=true;
	 for(int iwk=0;iwk<r;iwk++)indx.push_back(iwk);
	 while(done)
	 {
	    done=false;
	    for(int owk=0;owk<r;owk++)
	    {
	       //A new combination has been generated!
	       comb.Push(alpha[indx[owk]]);
	    }
	    /*
	       printf("New combination:\n");
	       for(int i=0;i<comb.Length();i++)
	       printf("%g ",comb[i]);
	       printf("\n");
	     */
	    //Calculate P(A|BC)
	    par.Copy(GetConditionalDist(score,cov,comb));
	    //printf("pars are:%g %g\n",par[0],par[1]);
	    double chisq,condProb,prob;
	    if(par[1]==0.0)
	    {
	       condProb = 0.0;
	    }
	    else 
	    {
	       chisq = (t_max-par[0])*(t_max-par[0])/par[1];
	       if(chisq<0.0)
		  chisq = -chisq;
	       condProb = chidist(chisq,1);
	    }
	    //printf("conditional prob is:%g\n",condProb);
	    String hashKey;

	    if(r==2)
	    {
	       int tmp = comb[0];
	       hashKey += tmp;
	       tmp = comb[1];
	       hashKey += tmp;
	       //	       printf("hashKey is: %s\n",hashKey.c_str());
	       prob = condProb*uni;	       
	       //prob = exp(log(condProb) + log(uni));
	       //printf("joint prob is:%g\n",prob);
	       jointProbHash.SetDouble(hashKey,prob);
	       hashKey.Clear();
	    }
	    else
	    {
	       for(int i=1;i<r;i++)
	       {
		  int tmp = comb[i];
		  hashKey += tmp;
	       }
	       //	       printf("hashKey is: %s\n",hashKey.c_str());
	       prob = jointProbHash.Double(hashKey);
	       //printf("joint prob is:%g\n",prob);
	       //if(prob == _NAN_)
	       //{
	       //	  printf("Error! Joint probability %s is not hashed in properly.\n",hashKey);
	       //     }
	       prob *= condProb; //This is the new joint prob.
	       String newKey;
	       int tmp = comb[0];
	       newKey += tmp;
	       newKey += hashKey;
	       //	       printf("newKey is: %s\n",newKey.c_str());
	       jointProbHash.SetDouble(newKey,prob);
	       hashKey.Clear();
	    }
	    pvalue -= prob;
	    //printf("pvalue so far is:%g\n",pvalue);
	    comb.Clear();
	    for(int iwk=r-1;iwk>=0;iwk--)
	    {
	       if(indx[iwk]<=(n-1)-(r-iwk))
	       {
		  indx[iwk]++;
		  for(int swk=iwk+1;swk<r;swk++)
		  {
		     indx[swk]=indx[swk-1]+1;
		  }
		  iwk=-1;
		  done=true;
	       }
	    }
	 }
	 indx.clear();
      }
   }
   // printf("pvalue is:%g\n",pvalue);
   return pvalue;
}

//Get P(A|BC) where A is the first element in comb. Comb contains index only.
Vector VT::GetConditionalDist(Vector & score,Matrix & cov,Vector & comb)
{
   Vector result;
   result.Dimension(2);
   result.Zero();

   //chop score vector and sigma matrix
   Vector mu2;
   Matrix sub_cov;
   int dim = comb.Length()-1;
   mu2.Dimension(dim);
   sub_cov.Dimension(dim,dim);
   for(int i=0;i<dim;i++)
   {
      int idx1 = comb[i+1];
      mu2[i] = score[idx1];
      for(int j=0;j<dim;j++)
      {
	 int idx2 = comb[j+1];
	 sub_cov[i][j] = cov[idx1][idx2];
      }
   }
   /*printf("sub_cov is:\n");
     for(int i=0;i<dim;i++)
     {
     for(int j=0;j<dim;j++)
     {
     printf("%g ",sub_cov[i][j]);
     }
     printf("\n");
     }*/

   Matrix inv;
   inv.Copy(sub_cov);
   SVD svd;
   svd.InvertInPlace(inv);
   Vector sigma12;
   sigma12.Dimension(dim);
   for(int i=0;i<dim;i++)
   {
      int idx1 = comb[0];
      int idx2 = comb[i+1];
      sigma12[i] = cov[idx1][idx2];
   }
   /*
      printf("sigma12 is:\n");
      for(int i=0;i<dim;i++)
      printf("%g ",sigma12[i]);
      printf("\n");
    */
   Vector tmp;
   tmp.Dimension(dim);
   tmp.Zero();
   for(int i=0;i<dim;i++)
   {
      tmp[i] += sigma12.InnerProduct(inv[i]);
   }

   //Calculate Mu and Sigma here
   result[0] = tmp.InnerProduct(mu2);
   result[1] = 1.0 - tmp.InnerProduct(sigma12); 

   if(result[1]<0)
      result[1] = fabs(result[1]);

   return result;
} 

void VT::Run(Pedigree & ped)
{
   if(!vcfAnnotated)
   {
      error("\nERROR! VCF file needs to be annotated in INFO field.\n");
   }

   ChopAndWrite(ped);
}

void VT::Run(Pedigree & ped, FastTransform & trans,FastFit & engine,String & traitName,GroupFromAnnotation & group)
{
   if(!vcfAnnotated)
   {
      error("\nERROR! VCF file needs to be annotated in INFO field.\n");
   }

   if(asymptotic || permuteBool)
   {
      //Fill in UDX, UD, and UDY
      //trans.FinalizeTransform(engine.sigma_g2Hat, engine.sigma_e2Hat);

      if(!VT::VTwtMB && !VT::VTwtCMC_binary && !VT::VTwtCMC_counts)
      {
	 VTwtCMC_binary=true;
	 printf("Reminder: weight function was specified for Variable-threshold test. CMC_binary weight will be use in the following VT tests.\n");
      }

      fixedEff.Dimension(trans.persons);
      //      beta_hat.Dimension(trans.fixedEffects);
      //    engine.EstimateAnalyticalBeta(trans.UDY,trans, beta_hat);
      for(int i=0;i<trans.persons;i++)
      {
	 fixedEff[i] = trans.UDX[i].InnerProduct(engine.betaHat);
      }
      residuals_original.Copy(trans.UDY);
      residuals_original.Subtract(fixedEff);
      sigma2_hat=1.0;
      //sigma2_hat = residuals_original.SumSquares()/trans.persons;

      reader.open(GroupFromAnnotation::vcfInput,header);
      reader.readVcfIndex();

      if(asymptotic)
      {
	 printf("\nStart doing VT tests using asmptotic method ...\n");
	 CalculateAsymptotic(ped,trans,engine,traitName,group);
      }
      if(permuteBool)
      { 
	 printf("\nStart doing VT tests using maximum of permutations %d using Besag Clifford stopping rule ...\n",permuteMax);
	 CalculatePermute(ped,trans,engine,traitName);
      }
   }
}
